use std::{borrow::Cow, path::PathBuf};

use ropey::Rope;

use crate::{history::History, marks::MarkSet, transaction::Transaction};

/// A path for an open text buffer.
///
/// This indicates where the text data of the buffer came from, and
/// where it should be saved to.
#[derive(Debug, Clone, Eq, PartialEq)]
pub enum BufferPath {
    File(PathBuf), // A buffer for a normal file on disk.
    Temp(usize),   // A temporary buffer, with a number ID.
}

/// An open text buffer, currently being edited.
#[derive(Debug, Clone)]
pub struct Buffer {
    pub path: BufferPath,
    pub edits_since_saved: i64, // Tracks whether this buffer is out of sync with the last save.
    pub text: Rope,             // The actual text content.
    pub mark_sets: Vec<MarkSet>, // MarkSets for cursors, view positions, etc.
    history: History,
}

impl Buffer {
    pub fn new(text: Rope, path: BufferPath) -> Buffer {
        Buffer {
            path: path,
            edits_since_saved: 0,
            text: text,
            mark_sets: Vec::new(),
            history: History::new(),
        }
    }

    pub fn is_dirty(&self) -> bool {
        self.edits_since_saved != 0
    }

    /// Saves the buffer to disk.
    pub fn save(&mut self) {
        self.edits_since_saved = 0;
        todo!();
    }

    /// Replaces the given range of chars with the given text.
    ///
    /// The range does not have to be ordered (i.e. the first component can be
    /// greater than the second).
    pub fn edit(&mut self, char_idx_range: (usize, usize), text: &str) {
        if self.edits_since_saved >= 0 {
            self.edits_since_saved = self.edits_since_saved.wrapping_add(1);
        } else {
            // The undo state matching the last save is inaccessible, so
            // set this to a similarly inaccessible number.
            self.edits_since_saved = i64::MAX;
        }

        // Get the range, properly ordered.
        let (start, end) = if char_idx_range.0 < char_idx_range.1 {
            (char_idx_range.0, char_idx_range.1)
        } else {
            (char_idx_range.1, char_idx_range.0)
        };

        // Build the transaction.
        let trans = Transaction::from_edit(
            self.text.char_to_byte(start),
            &Cow::from(self.text.slice(start..end)),
            text,
        );

        // Update mark sets.
        for mark_set in self.mark_sets.iter_mut() {
            trans.apply_to_marks(mark_set);
            mark_set.make_consistent();
        }

        // Apply the edit to the document.
        trans.apply(&mut self.text);

        // Update undo stack.
        self.history.push_edit(trans);
    }

    /// Un-does the last edit if there is one, and returns the
    /// Transaction that was inverse-applied, in case it's needed for
    /// further processing.
    ///
    /// Returns None if there is no edit to undo.
    pub fn undo(&mut self) -> Option<&Transaction> {
        if let Some(trans) = self.history.undo() {
            self.edits_since_saved = self.edits_since_saved.wrapping_sub(1);

            // Update mark sets.
            for mark_set in self.mark_sets.iter_mut() {
                trans.apply_inverse_to_marks(mark_set);
                mark_set.make_consistent();
            }

            // Apply the edit to the document.
            trans.apply_inverse(&mut self.text);

            return Some(trans);
        } else {
            return None;
        }
    }

    /// Re-does the last edit if there is one, and returns the
    /// Transaction that was applied, in case it's needed for further
    /// processing.
    ///
    /// Returns None if there is no edit to redo.
    pub fn redo(&mut self) -> Option<&Transaction> {
        if let Some(trans) = self.history.redo() {
            self.edits_since_saved = self.edits_since_saved.wrapping_add(1);

            // Update mark sets.
            for mark_set in self.mark_sets.iter_mut() {
                trans.apply_to_marks(mark_set);
                mark_set.make_consistent();
            }

            // Apply the edit to the document.
            trans.apply(&mut self.text);

            return Some(trans);
        } else {
            return None;
        }
    }

    /// Creates a new empty mark set, and returns the set index.
    pub fn add_mark_set(&mut self) -> usize {
        self.mark_sets.push(MarkSet::new());
        return self.mark_sets.len() - 1;
    }
}
